/*******************************************************************************
  File: BtreeIndex_Page.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
	The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "IndexedFile.h"


namespace Util {

// XDR description
//	Page {
//		unsigned int	num_keys;
//		unsigned int	prev_page;
//		unsigned int	next_page;
//		unsigned int	dummy;	// always 0
//		opaque			data[PageSize - 16]; -- PageSize minus the header variables
//	};



template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::BtreeIndex_Page(size_t pid)
  : _pageId(pid),
    _dataLen(0),
	_prevPage(-1),
	_nextPage(-1),
	_dirty(false),
	_keyList()
{
	// empty
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::~BtreeIndex_Page()
{
	// empty
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
bool
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::must_split() const
{
	return (_dataLen * sizeof(XDR_Char)) > (PageSize - HeaderSize);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
size_t
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::add_key(BtreeIndex& btree, 
														 const std::pair<const Key, size_t>& x)
{
	// Encode the <key,link> to see how long it is
	char buf[ (PageSize - HeaderSize) / 2 ];
	XOutDataBuf xstrm(btree._xstream, buf, sizeof(buf));

	KeyTraits::encode(xstrm, x.first) << x.second;
	if (xstrm.fail()) throw std::ios_base::failure("Key too large");
	
	KeyInfo info(x.first, x.second, short(xstrm.tellp()));
	
	// Now put the KeyInfo object into the right place in the list
	int slot = 0;
	KeyList::iterator kit = _keyList.begin();
	for (; kit != _keyList.end(); ++kit) {
		if ( btree._comp(x.first, (*kit).key) ) break;
		++slot;
	}
	kit = _keyList.insert(kit, info);
	_dataLen += info.xlen;
	_dirty = true;
	return slot;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
bool
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::remove_key(int slot)
{
	if (size_t(slot) >= _keyList.size()) throw std::ios_base::failure("Invalid slot for erase");
	bool rtn = _dataLen < (PageSize - HeaderSize) / (2 * sizeof(XDR_Char)); // less than half full already
	// Find the slot
	KeyList::iterator kit = _keyList.begin();
	while (slot > 0) {
		++kit;
		--slot;
	}
	_keyList.erase(kit);
	_dataLen -= (*kit).xlen;
	_dirty = true;
	return rtn;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::update_key(size_t slot, size_t link)
{
	if (slot >= _keyList.size()) throw std::ios_base::failure("Invalid slot for update");
	// Find the slot
	KeyList::iterator kit = _keyList.begin();
	while (slot > 0) {
		++kit;
		--slot;
	}
	(*kit).link = link;
	_dirty = true;
	return;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::insert_page_link(KeyList::iterator kit, Page* p)
{
	// Build a KeyInfo object from the link page
	KeyInfo info = p->key_list().front();
	info.link = p->page_id();	// doesn't change xlen
	_dataLen += info.xlen;
	_keyList.insert(kit, info);
	_dirty = true;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::update_page_link(Page* p)
{
	// The first key on the decendant page has changed
	// Find the page reference
	KeyList::iterator kit = _keyList.begin();
	for (; kit != _keyList.end(); ++kit) {
		if ((*kit).link == p->page_id()) break;
	}
	if (kit == _keyList.end()) throw std::ios_base::failure("Invalid page");
	// Carefully replace the keyinfo at this slot
	_dataLen -= (*kit).xlen;
	(*kit) = p->key_list().front();
	(*kit).link = p->page_id();
	_dataLen += (*kit).xlen;
	_dirty = true;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
bool
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::remove_page_link(Page* p)
{
	bool rtn = _dataLen > (PageSize - HeaderSize) / (2 * sizeof(XDR_Char));
	// Find the page reference
	KeyList::iterator kit = _keyList.begin();
	for (; kit != _keyList.end(); ++kit) {
		if ((*kit).link == p->page_id()) break;
	}
	if (kit == _keyList.end()) throw std::ios_base::failure("Invalid page");
	// Remove it
	_dataLen -= (*kit).xlen;
	_keyList.erase(kit);
	_dirty = true;
	return rtn;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>*
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::split(BtreeIndex& btree)
{
	// Go through the page and find where to split it
	KeyList::iterator kit = _keyList.begin();
	size_t len = 0;
	while (len < (PageSize - HeaderSize) / (2 * sizeof(XDR_Char))) {
		len += (*kit).xlen;
		++kit;
	}
	
	// Now move the rest of the elements to a new page
	Page* newp = btree.get_free_page();
	KeyList& kl = newp->_keyList;
	kl.splice(kl.end(), _keyList, kit, _keyList.end());
	newp->_dataLen = _dataLen - len;
	_dataLen = len;

	// link the page in after this one
	newp->next_page(_nextPage);
	newp->prev_page(_pageId);
	Page* nextp = btree.get_page(_nextPage);
	if (nextp) nextp->prev_page(newp->page_id());
	_nextPage = newp->page_id();

	// make sure both pages get flushed
	_dirty = newp->_dirty = true;

	// Return the new page
	return newp;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>*
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::consolidate(BtreeIndex& btree)
{
	// Get the prev and next pages - we will need them both
	Page* prevp = btree.get_page(_prevPage);
	Page* nextp = btree.get_page(_nextPage);

	// See if we can consolidate with the prev page
	if (prevp and prevp->_dataLen + _dataLen <= (PageSize - HeaderSize) / sizeof(XDR_Char)) {	// consolidate
		KeyList& kl = prevp->_keyList;
		kl.splice(kl.end(), _keyList);
		prevp->_dataLen += _dataLen;

		// make the current page free
		prevp->next_page(_nextPage);
		if (nextp)
			nextp->prev_page(_prevPage);
		btree.add_free_page(this);
		return this;
	}

	// See if we can consolidate with the next page
	if (nextp and nextp->_dataLen + _dataLen <= (PageSize - HeaderSize) / sizeof(XDR_Char)) {
		KeyList& kl = nextp->_keyList;
		_keyList.splice(_keyList.end(), kl);
		_dataLen += nextp->_dataLen;

		// free the next page
		Page* p = btree.get_page(nextp->_nextPage);
		if (p) {
			p->prev_page(_pageId);
			next_page(p->_pageId);
		}
		btree.add_free_page(nextp);
		return nextp;
	}
	
	// No consolodation possible
	return 0;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>*
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::shuffle(BtreeIndex& btree)
{
	// See if we can get a key from the previous page
	Page* prevp = btree.get_page(_prevPage);
	if (prevp) {
		KeyList& k1 = prevp->_keyList;
		_keyList.splice(_keyList.begin(), k1, --k1.end());
		prevp->_dataLen -= _keyList.front().xlen;
		_dataLen += _keyList.front().xlen;
		return this;
	}

	// Otherwise see if we can get a key from the next page
	Page* nextp = btree.get_page(_nextPage);
	if (nextp) {
		KeyList& kl = nextp->_keyList;
		_keyList.splice(_keyList.end(), kl, kl.begin());
		nextp->_dataLen -= _keyList.back().xlen;
		_dataLen += _keyList.back().xlen;
		return nextp;
	}
	
	// No shuffle possible
	return 0;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
oXDR_Stream&
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::encode(oXDR_Stream& xstrm)
{
	xstrm.seekp(_pageId);

	xstrm << _keyList.size();
	xstrm << _prevPage;
	xstrm << _nextPage;
	xstrm << 0;

	char data[PageSize - HeaderSize];
	XOutDataBuf xdatastrm(xstrm, data, sizeof(data));

	KeyList::iterator it = _keyList.begin(); 
	for (; it != _keyList.end(); ++it) {
		KeyTraits::encode(xdatastrm, (*it).key) << (*it).link;
	}
	xstrm.put_opaque(data, sizeof(data));

	_dirty = false;
	return xstrm;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
iXDR_Stream&
BtreeIndex_Page<Key,Compare,KeyTraits,PageSize>::decode(iXDR_Stream& xstrm)
{
	xstrm.seekg(_pageId);

	unsigned int numKeys;
	unsigned int dummy;
	xstrm >> numKeys;
	xstrm >> _prevPage;
	xstrm >> _nextPage;
	xstrm >> dummy;

	char data[PageSize - HeaderSize];
	xstrm.get_opaque(data, sizeof(data));

	XInDataBuf xdatastrm(xstrm, data, sizeof(data));

	_dataLen = 0;
	_keyList.clear();
	while (numKeys-- > 0) {
		KeyInfo info;
		info.xlen = short(xdatastrm.tellg()); // the start
		KeyTraits::decode(xdatastrm, info.key);
		xdatastrm >> info.link;
		info.xlen = short(xdatastrm.tellg()) - info.xlen;
		_dataLen += info.xlen;
		_keyList.push_back(info);
	}
	return xstrm;
}


} //> namespace
