/*******************************************************************************
  File: XIO.cpp
  Copyright (c) 1999 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, to prevent the 
  continued distribution of this Software under an open source distribution
  model. This includes, but is not limited to, such things as copyrighting
  or claiming authorship of the Software.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
  Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
#include "XIO.h"

namespace Util {


XIOFilebuf::XIOFilebuf()
{
	// set up the put buffer, the get buffer is empty
	setp(_obuf, _obuf + sizeof(_obuf) / sizeof(char_type));
}


XIOFilebuf::~XIOFilebuf()
{
	// empty
}


bool
XIOFilebuf::is_open()
{
	return _fb.is_open();
}

XIOFilebuf*
XIOFilebuf::open(const std::string& name, std::ios_base::openmode mode)
{
	return (_fb.open(name.c_str(), mode | std::ios_base::binary) ? this : 0);
}

XIOFilebuf*
XIOFilebuf::close()
{
	overflow(traits_type::eof());
	return (_fb.close() ? this : 0);
}


XIOFilebuf::inherited*
XIOFilebuf::setbuf(char* s, streamsize n)
{
	// do nothing
	return this;
}


XIOFilebuf::pos_type
XIOFilebuf::seekoff(off_type off, std::ios_base::seekdir dir, std::ios_base::openmode which)
{
	// sync the buffers
	sync();

	// clear the get area
	setg(null, null, null);
	
	// adjust the offset by the size of an XDR_Char (which better be 4)
	return _fb.pubseekoff((off * sizeof(char_type)), dir, which) / sizeof(char_type);
}


XIOFilebuf::pos_type
XIOFilebuf::seekpos(pos_type pos, std::ios_base::openmode which)
{
	// sync the buffers
	sync();

	// clear the get area
	setg(null, null, null);
	
	// adjust the position by the size of an XDR_Char.
	return _fb.pubseekpos((pos * sizeof(char_type)), which) / sizeof(char_type);
}


int
XIOFilebuf::sync()
{
	// write any output to the controled sequence
	streamsize n = pptr() - pbase();
	if (n > 0) {
		streamsize x = _fb.sputn(reinterpret_cast<char*>(_obuf), (n * sizeof(char_type)));
		if (x < 0 or size_t(x) != (n * sizeof(char_type))) return -1;
		pbump(-n);
	}
	return _fb.pubsync();
}


XIOFilebuf::int_type
XIOFilebuf::underflow()
{
	// read some characters into the get buffer
	streamsize x = _fb.sgetn(reinterpret_cast<char*>(_ibuf), sizeof(_ibuf));
	if (x == 0) {
		return traits_type::eof();
	}
	if (x % 4 != 0) {
		throw std::ios::failure("Unexpected end-of-file in middle of XDR_Char");
	}
	
	// set up the get area
	setg(_ibuf, _ibuf, _ibuf + (x / sizeof(char_type)));
	return traits_type::to_int_type(*_ibuf);
}


XIOFilebuf::int_type
XIOFilebuf::overflow(int_type c)
{
	// write the buffer
	int rtn = sync();
	if (rtn == -1)
		throw std::ios::failure("Error writing output to base stream");
	
	// put the new character in the buffer
	if (not traits_type::eq_int_type(c, traits_type::eof())) {
		*_obuf = traits_type::to_char_type(c);
		pbump(1);
	}
	return traits_type::not_eof(c);
}


//
// input file stream
//
iXIOFileStream::iXIOFileStream()
{
	init(&_xfb);
}


iXIOFileStream::iXIOFileStream(const std::string& name, std::ios_base::openmode mode)
{
	init(&_xfb);
	open(name, mode);
}

iXIOFileStream::~iXIOFileStream()
{
	// empty
}

bool
iXIOFileStream::is_open()
{
	return rdbuf()->is_open();
}

void
iXIOFileStream::open(const std::string& name, std::ios_base::openmode mode)
{
	if (not rdbuf()->open(name, mode)) setstate(std::ios_base::badbit);
}

void
iXIOFileStream::close()
{
	if (not rdbuf()->close()) setstate(std::ios_base::badbit);
}


//
// output file stream
//
oXIOFileStream::oXIOFileStream()
{
	init(&_xfb);
}

oXIOFileStream::oXIOFileStream(const std::string& name, std::ios_base::openmode mode)
{
	init(&_xfb);
	open(name, mode);
}

oXIOFileStream::~oXIOFileStream()
{
	// empty
}

bool
oXIOFileStream::is_open()
{
	return rdbuf()->is_open();
}

void
oXIOFileStream::open(const std::string& name, std::ios_base::openmode mode)
{
	if (not rdbuf()->open(name, mode)) setstate(std::ios_base::badbit);
}

void
oXIOFileStream::close()
{
	if (not rdbuf()->close()) setstate(std::ios_base::badbit);
}


//
// input / output file streams
//
XIOFileStream::XIOFileStream()
{
	init(&_xfb);
}

XIOFileStream::XIOFileStream(const std::string& name, std::ios_base::openmode mode)
{
	init(&_xfb);
	open(name, mode);
}


XIOFileStream::~XIOFileStream()
{
	// empty
}

bool
XIOFileStream::is_open()
{
	return rdbuf()->is_open();
}

void
XIOFileStream::open(const std::string& name, std::ios_base::openmode mode)
{
	if (not rdbuf()->open(name, mode)) setstate(std::ios_base::badbit);
}

void
XIOFileStream::close()
{
	if (not rdbuf()->close()) setstate(std::ios_base::badbit);
}


//
// String based stream
//
XIOStringbuf::XIOStringbuf(std::ios_base::openmode mode)
  : _sb(mode | std::ios_base::binary)
{
	// set up the put buffer, the get buffer is empty
	setp(_obuf, _obuf + sizeof(_obuf) / sizeof(char_type));
}

XIOStringbuf::XIOStringbuf(const std::string& str, std::ios_base::openmode mode)
  : _sb(str, mode | std::ios_base::binary)
{
	// set up the put buffer, the get buffer is empty
	setp(_obuf, _obuf + sizeof(_obuf) / sizeof(char_type));
}

XIOStringbuf::~XIOStringbuf()
{
	// empty
}

std::string
XIOStringbuf::str() const
{
	return _sb.str();
}

void
XIOStringbuf::str(const std::string& s)
{
	_sb.str(s);
}


XIOStringbuf::inherited*
XIOStringbuf::setbuf(char* s, streamsize n)
{
	// do nothing
	return this;
}


XIOStringbuf::pos_type
XIOStringbuf::seekoff(off_type off, std::ios_base::seekdir dir, std::ios_base::openmode which)
{
	// sync the buffers
	sync();

	// clear the get area
	setg(null, null, null);
	
	// adjust the offset by the size of an XDR_Char (which better be 4)
	return _sb.pubseekoff((off * sizeof(char_type)), dir, which) / sizeof(char_type);
}


XIOStringbuf::pos_type
XIOStringbuf::seekpos(pos_type pos, std::ios_base::openmode which)
{
	// sync the buffers
	sync();

	// clear the get area
	setg(null, null, null);
	
	// adjust the position by the size of an XDR_Char.
	return _sb.pubseekpos((pos * sizeof(char_type)), which) / sizeof(char_type);
}


int
XIOStringbuf::sync()
{
	// write any output to the controled sequence
	streamsize n = pptr() - pbase();
	if (n > 0) {
		streamsize x = _sb.sputn(reinterpret_cast<char*>(_obuf), (n * sizeof(char_type)));
		if (x < 0 or size_t(x) != (n * sizeof(char_type))) return -1;
		pbump(-n);
	}
	return _sb.pubsync();
}


XIOStringbuf::int_type
XIOStringbuf::underflow()
{
	// read some characters into the get buffer
	streamsize x = _sb.sgetn(reinterpret_cast<char*>(_ibuf), sizeof(_ibuf));
	if (x == 0) {
		return traits_type::eof();
	}
	if (x % 4 != 0) {
		throw std::ios::failure("Unexpected end-of-file in middle of XDR_Char");
	}
	
	// set up the get area
	setg(_ibuf, _ibuf, _ibuf + (x / sizeof(char_type)));
	return traits_type::to_int_type(*_ibuf);
}


XIOStringbuf::int_type
XIOStringbuf::overflow(int_type c)
{
	// write the buffer
	int rtn = sync();
	if (rtn == -1)
		throw std::ios::failure("Error writing output to base stream");
	
	// put the new character in the buffer
	*_obuf = traits_type::to_char_type(c);
	pbump(1);
	return traits_type::not_eof(c);
}


XIOStringStream::XIOStringStream(std::ios_base::openmode mode)
  : _xsb(mode)
{
	init(&_xsb);
}


XIOStringStream::XIOStringStream(const std::string& str, std::ios_base::openmode mode)
  : _xsb(str, mode)
{
	init(&_xsb);
}

XIOStringStream::~XIOStringStream()
{
	// empty
}

std::string
XIOStringStream::str() const
{
	return rdbuf()->str();
}

void
XIOStringStream::str(const std::string& s)
{
	rdbuf()->str(s);
}


} //> namespace Util
