/*******************************************************************************
  File: check.h
  Copyright (c) 1999 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, to prevent the 
  continued distribution of this Software under an open source distribution
  model. This includes, but is not limited to, such things as copyrighting
  or claiming authorship of the Software.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
$Header: $

//!-----------------------------------------------------------------------------
Description:
  Provides facilities to assist in error checking.
  There are two sections: assertions and checks.

  Assertions --
  This header provides a number of "assert" macros. These are in addition to 
  the standard "assert" macro defined in <assert.h>. These macros throw 
  exceptions, but are otherwise identical to the regular 'assert' macro. 
  The macros are:
    assert_logic    - throws "std::logic_error"
    assert_range    - throws "std::out_of_range"
    assert_length   - throws "std::length_error"
    assert_domain   - throws "std::domain_error"
    assert_argument - throws "std::invalid_argument"
  
  Besides defining these macros, the implementation file also provides 
  alternate implementation of the 'assert helper' function that is typically
  called by the standard 'assert' macro. This new implementation throws a
  logic_error exception instead of the "traditional" behavior.

  Like the standard 'assert' macro, these macros are disabled when the 
  preprocessor symbol NDEBUG is defined. In addition, each new 'assert' macro
  has its own flag which can be used to disable it alone.

  Checks --
  These are a series of macros similar to the 'assert_xxx' macros above, but
  with two differences:
    a. These macros can not be disabled like the assert macros can
    b. The check macros take a second parameter. This is usually a string
    that is the name of the function. It can be anything that can be converted
    to a string.
  
  This header includes "raise.h" so that the raise_handler is available to
  anyone who also includes check.h.

  This header also defines the template 'down_cast' which is controlled by
  NDEBUG. It is inside the multiple-include guard, so it will not cause
  multiple definitions if the header is included more than once.

//#-----------------------------------------------------------------------------
Notes:
  -- The helper functions used by these macros (raise_xxx) are declared at the
  the global namespace. This allows the functions to be used in the macros 
  without namespace qualification. This, in turn, allows some user of the macros 
  to provide local scope versions of the helper functions and still use the 
  macros.

  -- Parts of this header are outside the multi-include guard. This is to mimic 
  the functionallity of the Standard C header 'assert.h'. In other words, this
  header can be included multiple times, and the result depends upon the
  state of the various preprocessor symbols:
  NO_RANGE_ASSERT -- disables the 'assert_range' macro
  NO_LENGTH_ASSERT -- disables the 'assert_length' macro
  NO_DOMAIN_ASSERT -- disables the 'assert_domain' macro
  NO_ARGUMENT_ASSERT -- disables the 'assert_argument' macro

  -- Defining NDEBUG disables all the assert_xxx macros,

//%****************************************************************************/
#ifndef UTIL_CHECK_H
#define UTIL_CHECK_H

// Platform
#include <platform.h>

// Utility headers
#include "raise.h"  // not used here but included for clients

// Standard C++ headers
#include <string>

//
// Helper functions
//
void raise_logic_error(const std::string& why, const std::string& where = "");
void raise_domain_error(const std::string& why, const std::string& where = "");
void raise_out_of_range(const std::string& why, const std::string& where = "");
void raise_length_error(const std::string& why, const std::string& where = "");
void raise_invalid_argument(const std::string& why, const std::string& where = "");

// These helper functions do not have corresponding macros, but are 
// potentially useful by themselves.
void raise_runtime_error(const std::string& why, const std::string& where = "");
void raise_range_error(const std::string& why, const std::string& where = "");
void raise_overflow_error(const std::string& why, const std::string& where = "");
void raise_underflow_error(const std::string& why, const std::string& where = "");

//
// Check macros
//
#define check(EXPR, FUNC) \
    ((EXPR) ? (void)0 \
            : raise_logic_error("Check failed(" #EXPR ")", FUNC))

#define check_logic(EXPR, FUNC) \
    ((EXPR) ? (void)0 \
            : raise_logic_error("Check failed(" #EXPR ")", FUNC))

#define check_domain(EXPR, FUNC) \
    ((EXPR) ? (void)0 \
            : raise_domain_error("Check failed(" #EXPR ")", FUNC))

#define check_range(EXPR, FUNC) \
    ((EXPR) ? (void)0 \
            : raise_out_of_range("Check failed(" #EXPR ")", FUNC))

#define check_length(EXPR, FUNC) \
    ((EXPR) ? (void)0 \
            : raise_length_error("Check failed(" #EXPR ")", FUNC))

#define check_argument(EXPR, FUNC) \
    ((EXPR) ? (void)0 \
            : raise_invalid_argument("Check failed(" #EXPR ")", FUNC))


//
// helper function used by the assert macros
//
std::string make_where_str(const char* file, int lineno);

// helper macro
#define HERE \
  (make_where_str(__FILE__, __LINE__))

// Downcast template - non-debugging version does not throw
#ifndef NDEBUG

template<typename X, typename Y>
inline X down_cast(Y y)
{ return dynamic_cast<X>(y); }

#else 

template<typename X, typename Y>
inline X down_cast(Y y)
{ return static_cast<X>(y); }

#endif

#endif // UTIL_CHECK_H


//
// assert macros
//  Note these are outside the include guard
//
#undef assert_logic
#undef assert_range
#undef assert_length
#undef assert_domain
#undef assert_argument

#ifndef NDEBUG

#ifndef NO_LOGIC_ASSERT
  #define assert_logic(EXPR) \
    ((EXPR) ? (void)0 \
            : raise_logic_error("Assertion failed(" #EXPR ")", HERE))
#else
  #define assert_logic(EXPR) ((void)0)
#endif

#ifndef NO_RANGE_ASSERT
  #define assert_range(EXPR) \
    ((EXPR) ? (void)0 \
            : raise_out_of_range("Assertion failed(" #EXPR ")", HERE))
#else
  #define assert_range(EXPR) ((void)0)
#endif
  
#ifndef NO_LENGTH_ASSERT
  #define assert_length(EXPR) \
    ((EXPR) ? (void)0 \
            : raise_length_error("Assertion failed(" #EXPR ")", HERE))
#else
  #define assert_length(EXPR) ((void)0)
#endif
  
#ifndef NO_DOMAIN_ASSERT
  #define assert_domain(EXPR) \
    ((EXPR) ? (void)0 \
            : raise_domain_error("Assertion failed(" #EXPR ")", HERE))
#else
  #define assert_domain(EXPR) ((void)0)
#endif
  
#ifndef NO_ARGUMENT_ASSERT
  #define assert_argument(EXPR) \
    ((EXPR) ? (void)0 \
            : raise_invalid_argument("Assertion failed(" #EXPR ")", HERE))
#else
  #define assert_argument(EXPR) ((void)0)
#endif

#else

  #define assert_logic(EXPR)    ((void)0)
  #define assert_range(EXPR)    ((void)0)
  #define assert_length(EXPR)   ((void)0)
  #define assert_domain(EXPR)   ((void)0)
  #define assert_argument(EXPR) ((void)0)

#endif
