/*******************************************************************************
  File: BtreeIndex.h
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
#ifndef UTIL_BTREE_INDEX_H
#define UTIL_BTREE_INDEX_H

// Platform header
#include "platform.h"

// Utility headers
#include "raise.h"

// XDR Stream headers
#include "XDRStream.h"
#include "XIO.h"


// Standard C++ headers
#include <functional>
#include <string>
#include <ios>
#include <list>

namespace Util {

//
// XDR_Traits --
//
template <typename Item>
struct XDR_Traits {
	typedef		XIOFileStream			xio_type;
	static oXDR_Stream& encode(oXDR_Stream& oxs, const Item& val)
		{ return oxs << val; }
	static iXDR_Stream& decode(iXDR_Stream& ixs, Item& val)
		{ return ixs >> val; }
};


//
// BtreeIndex --
//
template<typename Key,
		 typename Compare = std::less<Key>,
		 typename KeyTraits = XDR_Traits<Key>,
		 size_t PageSize = 4096>
class BtreeIndex {
public:
	// public types
	class Iterator;
	class StreamposProxy;

	typedef Key							key_type;
	typedef std::pair<const Key, size_t> value_type;
	typedef Compare						key_compare;
	typedef Iterator					iterator;
	typedef size_t						size_type;
	typedef ptrdiff_t					difference_type;

	enum DeleteStrategy { None, Compact, Shuffle };
	enum UpdateStrategy { Unsafe, MostlySafe, Safe };
	
	class StreamposProxy {
		friend class BtreeIndex;
		// data
		size_t			_link;
		BtreeIndex*		_btree;
		size_t			_page;
		size_t			_slot;
	public:
		// LWC - default copy and destroy Ok
		// conversion
		operator size_t();
		// assignment
		void operator=(size_t newlink);
	private:
		StreamposProxy(size_t link, 
					   BtreeIndex* btree, 
					   size_t page,
					   size_t slot);
	};


	class Iterator : public std::iterator<std::bidirectional_iterator_tag, 
										  std::pair<const Key, StreamposProxy> > {
		friend class BtreeIndex;
		// data
		BtreeIndex*			_btree;
		size_t				_page;
		size_t				_slot;
	public:
		Iterator()	{}

		std::pair<const Key, StreamposProxy> operator*();
		Iterator&					operator++();
		Iterator					operator++(int);
		Iterator&					operator--();
		Iterator					operator--(int);

		bool operator==(const Iterator& rhs) const;
		bool operator!=(const Iterator& rhs) const;
	private:
		Iterator(BtreeIndex* btree, size_t page, size_t slot);
	};


	// construct / destroy
	explicit BtreeIndex(Compare comp = Compare());
	explicit BtreeIndex(const std::string& name, 
							std::ios_base::openmode mode = std::ios_base::in|std::ios_base::out,
							Compare comp = Compare());
	~BtreeIndex();

	// open/close/test state
	bool			is_open();
	void			open(const std::string& name, std::ios_base::openmode mode = std::ios_base::in|std::ios_base::out);
	void			close();
	void			flush();
	operator void*() const;
	bool operator!() const;

	// iterators
	iterator		begin();
	iterator		end();

	// capacity
	bool			empty();
	size_type		size();

	// modifiers
	iterator		insert(const value_type& x);
	iterator		insert(const Key& k);
	template<typename InputIterator>
		void		insert(InputIterator first, InputIterator last);

	void			erase(Iterator pos);
	size_type		erase(const Key& x);
	void			erase(iterator first, iterator last);
	void			clear();

	// observers
	key_compare		key_comp() const;
	size_type		page_size() const;
	XDR_Stream&		xdrstream();

	// operations
	iterator		find(const Key& x);
	size_type		count(const Key& x);

	iterator		lower_bound(const Key& x);
	iterator		upper_bound(const Key& x);
	std::pair<iterator, iterator>
					equal_range(const Key& x);

	// utilities
	void			reindex();
	DeleteStrategy	set_delete_strategy(DeleteStrategy strategy);
	UpdateStrategy	set_update_strategy(UpdateStrategy strategy);
	int				set_cache_size(int size);

private:
	// private types
	class XIODataStreambuf;
	class XInDataBuf;
	class XOutDataBuf;
	class Page;
	class RootPage;

	class PageCache {
		// types
		typedef std::list<Page*> PageList;

		// data
		XDR_Stream&	_xstream;
		size_t		_minSize;
		size_t		_maxSize;
		PageList	_pageList;
	public:
		PageCache(XDR_Stream&);
		~PageCache();

		Page*		get_page(size_t page);
		void		add_page(Page* p); // puts a new page into the cache
		void		reset();	// deletes any excess pages
		void		clear();	// deletes all pages
		void		flush();	// writes any dirty pages out

		void		min_size(size_t size)	{ _minSize = size; }
		int			max_size()				{ return _maxSize; }
		void		max_size(size_t size)	{ _minSize = size; }
	}; //> PageCache

	// friends
	friend class Iterator;
	friend class StreamposProxy;
	friend class Page;
	friend class RootPage;

	// data
	Compare					_comp;
	typename KeyTraits::xio_type _xstream;
	RootPage*				_root;
	PageCache				_pageCache;
	DeleteStrategy			_dStrategy;
	UpdateStrategy			_uStrategy;

	// Iterator utility functions - used by friends
	std::pair<const Key, StreamposProxy>
					get_value(size_t page, size_t slot);
	void			prev_iter(Iterator& it);
	void			next_iter(Iterator& it);

	// StreamposProxy utility functions - used by friends
	void			update_link(size_t page, size_t offset, size_t link);

	// BtreeIndex utility functions
	Page*			get_page(size_t pageId);
	Page*			get_free_page();
	void			add_free_page(Page* p);
	iterator		lower_bound(Page* p, int depth, const Key& k);
	iterator		upper_bound(Page* p, int depth, const Key& k);
	iterator		insert(Page*, int depth, const value_type& x);
	std::pair<bool,bool> erase(Page* p, int depth, const Key& k, iterator pos);

	void			flush(UpdateStrategy s);
}; //> BtreeIndex


//
// BtreeIndex::Page
//
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
class BtreeIndex<Key,Compare,KeyTraits,PageSize>::Page {
public:
	struct KeyInfo {
		Key			key;
		size_t		link;
		short		xlen;	// length of encoded <key,link> in data area
		KeyInfo(Key k, size_t l, short xl) : key(k), link(l), xlen(xl) {}
		KeyInfo() {}
	};
	typedef std::list<KeyInfo>	KeyList;

protected:
	// data
	const size_t _pageId;
	size_t		_dataLen;
	size_t		_prevPage;
	size_t		_nextPage;
	bool		_dirty;		// set when changed; cleared by encode
	KeyList		_keyList;

	enum { HeaderSize = 16 };

public:
	Page(size_t pageId);
	virtual ~Page();

	// access
	size_t		page_id() const			{ return _pageId; }
	size_t		data_len() const		{ return _dataLen; }
	size_t		prev_page() const		{ return _prevPage; }
	size_t		next_page() const		{ return _nextPage; }
	bool		is_dirty() const		{ return _dirty; }
	size_t		num_keys() const		{ return _keyList.size(); }

	// update
	KeyList&	key_list()				{ return _keyList; }

	void		data_len(size_t len)	{ _dataLen = len; _dirty = true; }
	void		prev_page(size_t pid)	{ _prevPage = pid; _dirty = true; }
	void		next_page(size_t pid)	{ _nextPage = pid; _dirty = true; }

	// test
	bool		must_split() const;

	// access / update the key list
	size_t		add_key(BtreeIndex&, const std::pair<const Key, size_t>&);  // returns the slot of the new key

	bool		remove_key(int slot);	// returns true if page less than 1/2 full
	void		update_key(size_t slot, size_t link);

	void		insert_page_link(typename KeyList::iterator kit, Page* p);
	void		update_page_link(Page* p);
	bool		remove_page_link(Page* p);  // returns same as remove_key

	// virtual functions
	virtual Page*		split(BtreeIndex&);  // returns the new page
	virtual Page*		consolidate(BtreeIndex&);	// returns the free page; or null
	virtual Page*		shuffle(BtreeIndex&);		// returns the page with the new first key
	virtual oXDR_Stream& encode(oXDR_Stream&);
	virtual iXDR_Stream& decode(iXDR_Stream&);
private:
	Page(const Page&);
	Page& operator=(const Page&);
}; //> BtreeIndex::Page


//
// BtreeIndex::RootPage
//
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
class BtreeIndex<Key,Compare,KeyTraits,PageSize>::RootPage : public Page {
	// types
	typedef typename Page::KeyInfo		KeyInfo;
	typedef typename Page::KeyList		KeyList;
	
	// data
	size_t		_totalKeys;
	short		_depth;
	bool		_dirtyTotal;
public:
	RootPage();
	virtual ~RootPage();

	// access
	size_t		total_keys() const		{ return _totalKeys; }
	short		depth() const			{ return _depth; }
	size_t		next_free_page() const	{ return _nextPage; }
	bool		is_total_dirty() const	{ return _dirtyTotal; }

	// update
	void		incr_total()			{ ++_totalKeys; _dirtyTotal = true; }
	void		decr_total()			{ --_totalKeys; _dirtyTotal = true; }
	void		add_free_page(Page* p)	{ p->next_page(next_page()); next_page(p->page_id()); }
	void		rem_free_page(Page* p)	{ next_page(p->next_page()); }

	// virtual functions
	virtual Page*		split(BtreeIndex&);
	virtual Page*		consolidate(BtreeIndex&);
	virtual oXDR_Stream& encode(oXDR_Stream&);
	virtual iXDR_Stream& decode(iXDR_Stream&);
}; //> BtreeIndex::RootPage


// class which handles read/write to/from an externally specified memory area
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
class BtreeIndex<Key,Compare,KeyTraits,PageSize>::XIODataStreambuf : public XDR_Streambuf {
	// data
	XDR_Char*		_buf;
	size_t			_size;
	size_t			_len;
public:
	XIODataStreambuf(char* buf, size_t size, size_t len = 0);
	~XIODataStreambuf();

	virtual pos_type seekoff(off_type off, 
	                         std::ios_base::seekdir dir, 
							 std::ios_base::openmode which);
	virtual pos_type seekpos(pos_type pos, std::ios_base::openmode which);
};


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
class BtreeIndex<Key,Compare,KeyTraits,PageSize>::XOutDataBuf : public oXDR_Stream {
	// data
	oXDR_Stream&		_refstrm;
	XIODataStreambuf	_xstreambuf;

public:
	XOutDataBuf(oXDR_Stream&, char* buf, size_t size);
	~XOutDataBuf();
};


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
class BtreeIndex<Key,Compare,KeyTraits,PageSize>::XInDataBuf : public iXDR_Stream {
	// data
	iXDR_Stream&		_refstrm;
	XIODataStreambuf	_xstreambuf;

public:
	XInDataBuf(iXDR_Stream&, char* buf, size_t len);
	~XInDataBuf();
};


//
// BtreeIndex::StreamposProxy - inline function definitions
//
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline BtreeIndex<Key,Compare,KeyTraits,PageSize>::StreamposProxy::operator size_t()
{ return _link; }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::StreamposProxy::operator=(size_t newlink)
{ _btree->update_link(_page, _slot, newlink); }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline BtreeIndex<Key,Compare,KeyTraits,PageSize>::StreamposProxy::StreamposProxy(
			size_t link, 
			BtreeIndex* btree, 
			size_t page, 
			size_t slot)
  : _link(link), _btree(btree), _page(page), _slot(slot) {}


//
// BtreeIndex::Iterator - inline function definitions
//
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline std::pair<const Key, typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::StreamposProxy>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::operator*()
{ return _btree->get_value(_page, _slot); }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator&
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::operator++()
{ _btree->next_iter(*this); return *this; }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::operator++(int)
{ Iterator t = *this; _btree->next_iter(*this); return t; }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator&
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::operator--()
{ _btree->prev_iter(*this); return *this; }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::operator--(int)
{ Iterator t = *this; _btree->prev_iter(*this); return t; }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline bool
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::operator==(const Iterator& rhs) const
{ return _btree == rhs._btree 
			and _page == rhs._page 
				and _slot == rhs._slot; }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline bool 
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::operator!=(const Iterator& rhs) const
{ return !(*this == rhs); }

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator::Iterator(
			BtreeIndex* btree, 
			size_t page, 
			size_t slot)
  : _btree(btree), _page(page), _slot(slot) {}


//
// BtreeIndex inline function definitions
//
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline bool
BtreeIndex<Key,Compare,KeyTraits,PageSize>::is_open()
{ return _xstream.is_open(); }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline 
BtreeIndex<Key,Compare,KeyTraits,PageSize>::operator void*() const
{
	if (_xstream.fail()) return null;
	return const_cast<BtreeIndex*>(this);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline bool
BtreeIndex<Key,Compare,KeyTraits,PageSize>::operator!() const
{ return _xstream.fail(); }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline bool
BtreeIndex<Key,Compare,KeyTraits,PageSize>::empty()
{ return size() == 0; }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline size_t
BtreeIndex<Key,Compare,KeyTraits,PageSize>::size()
{ return _root->total_keys(); }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::Iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::insert(const Key& key)
{ return insert(value_type(key, 0)); }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline Compare
BtreeIndex<Key,Compare,KeyTraits,PageSize>::key_comp() const
{ return _comp; }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline size_t
BtreeIndex<Key,Compare,KeyTraits,PageSize>::page_size() const
{ return PageSize; }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline XDR_Stream&
BtreeIndex<Key,Compare,KeyTraits,PageSize>::xdrstream()
{ return _xstream; }


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::DeleteStrategy
BtreeIndex<Key,Compare,KeyTraits,PageSize>::set_delete_strategy(DeleteStrategy s)
{
	DeleteStrategy t = _dStrategy;
	_dStrategy = s;
	return t;
}

template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::UpdateStrategy
BtreeIndex<Key,Compare,KeyTraits,PageSize>::set_update_strategy(UpdateStrategy s)
{
	UpdateStrategy t = _uStrategy;
	_uStrategy = s;
	return t;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
inline int
BtreeIndex<Key,Compare,KeyTraits,PageSize>::set_cache_size(int size)
{
	int t = _pageCache.max_size();
	_pageCache.max_size(size);
	return t;
}


} //> namespace


#ifdef EXPORT_NOT_SUPPORTED
#include "BtreeIndex.hxx"
#include "BtreeIndex_Page.hxx"
#include "BtreeIndex_RootPage.hxx"
#include "BtreeIndex_PageCache.hxx"
#include "BtreeIndex_XIODataBuf.hxx"
#endif


#endif
