/*******************************************************************************
  File: BtreeIndex_RootPage.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
	The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "IndexedFile.h"


namespace Util {

// XDR description
//	RootPage {
//		unsigned int	depth / page_size; -- 2 short ints combined
//		unsigned int	total_keys;
//		unsigned int	num_keys;
//		unsigned int	free_page;
//		opaque			data[PageSize - 16]; -- PageSize minus the header variables
//	};


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::RootPage::RootPage()
  : Page(0),
    _totalKeys(0),
	_depth(0),
	_dirtyTotal(false)
{
	// empty
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::RootPage::~RootPage()
{
	// empty
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::Page*
BtreeIndex<Key,Compare,KeyTraits,PageSize>::RootPage::split(BtreeIndex& btree)
{
	// Find where to split the page
	typename KeyList::iterator kit = _keyList.begin();
	size_t len = 0;
	while (len < (PageSize - HeaderSize) / (2 * sizeof(XDR_Char))) {
		len += (*kit).xlen;
		++kit;
	}

	// Spliting the root page means creating two new pages
	Page* newp1 = btree.get_free_page();
	Page* newp2 = btree.get_free_page();

	// move the first part to new page 1
	KeyList& kl1 = newp1->key_list();
	kl1.splice(kl1.end(), _keyList, _keyList.begin(), kit);
	newp1->data_len(len);

	// move whats left to the second new page
	KeyList& kl2 = newp2->key_list();
	kl2.splice(kl2.end(), _keyList);
	newp2->data_len(_dataLen - len);
	_dataLen = 0;

	// link the two pages together
	newp1->next_page(newp2->page_id());
	newp2->prev_page(newp1->page_id());

	// put the page links into the now empty root page
	insert_page_link(_keyList.end(), newp1);
	insert_page_link(_keyList.end(), newp2);
	_depth += 1;
	_dirty = true;
	return 0;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
typename BtreeIndex<Key,Compare,KeyTraits,PageSize>::Page*
BtreeIndex<Key,Compare,KeyTraits,PageSize>::RootPage::consolidate(BtreeIndex& btree)
{
	// consolodating the root page means eliminating the lower level page
	Page* oldp = btree.get_page(_keyList.front().link);
	_keyList.clear();

	// move the keys from the lower level page
	_keyList.swap(oldp->key_list());
	_dataLen = oldp->data_len();
	
	btree.add_free_page(oldp);
	_depth -= 1;
	_dirty = true;
	return oldp;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
oXDR_Stream&
BtreeIndex<Key,Compare,KeyTraits,PageSize>::RootPage::encode(oXDR_Stream& xstrm)
{
	xstrm.seekp(_pageId);

	unsigned long depth_pageSize = (_depth << 16) + PageSize;
	xstrm << depth_pageSize;
	xstrm << _totalKeys;
	xstrm << _keyList.size();	// num_keys;
	xstrm << _nextPage;			// free page id

	char data[PageSize - HeaderSize];
	XOutDataBuf xdatastrm(xstrm, data, sizeof(data));

	typename KeyList::iterator it = _keyList.begin(); 
	for (; it != _keyList.end(); ++it) {
		KeyTraits::encode(xdatastrm, (*it).key) << (*it).link;
	}
	xstrm.put_opaque(data, sizeof(data));

	_dirty = _dirtyTotal = false;
	return xstrm;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
iXDR_Stream&
BtreeIndex<Key,Compare,KeyTraits,PageSize>::RootPage::decode(iXDR_Stream& xstrm)
{
	xstrm.seekg(_pageId);

	unsigned int depth_pageSize;
	xstrm >> depth_pageSize;
	_depth = depth_pageSize >> 16;

	xstrm >> _totalKeys;
	unsigned int numKeys;

	xstrm >> numKeys;
	xstrm >> _nextPage;

	char data[PageSize - HeaderSize];
	xstrm.get_opaque(data, sizeof(data));

	XInDataBuf xdatastrm(xstrm, data, sizeof(data));

	_dataLen = 0;
	_keyList.clear();
	while (numKeys-- > 0) {
		KeyInfo info;
		info.xlen = short(xdatastrm.tellg());
		KeyTraits::decode(xdatastrm, info.key);
		xdatastrm >> info.link;
		info.xlen = short(xdatastrm.tellg()) - info.xlen;
		_dataLen += info.xlen;
		_keyList.push_back(info);
	}
	return xstrm;
}

} //> namespace

