/*******************************************************************************
  File: IndexedFile.h
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
#ifndef UTIL_INDEXED_FILE_H
#define UTIL_INDEXED_FILE_H

// Platform header
#include "platform.h"

// Index header
#include "BtreeIndex.h"

// Utility headers
#include "raise.h"

// XDR Stream headers
#include "XDRStream.h"
#include "XIO.h"


// Standard C++ headers
#include <functional>
#include <string>
#include <ios>
#include <list>

namespace Util {


//
// Class IndexedFile
//
template<typename Key,
         typename Item,
		 typename Compare = std::less<Key>,
		 typename KeyTraits = XDR_Traits<Key>,
		 typename ItemTraits = XDR_Traits<Item>,
		 size_t PageSize = 4096>
class IndexedFile {
public:
	// types
	class Iterator;

	typedef Key							key_type;
	typedef std::pair<const Key, Item>	value_type;
	typedef Compare						key_compare;
	typedef Iterator					iterator;
	typedef size_t						size_type;
	typedef ptrdiff_t					difference_type;
	typedef BtreeIndex<Key, Compare, KeyTraits, PageSize> index_type;

	class ItemProxy {
		friend class Iterator;
		// data
		IndexedFile*						_ifile;
		typename index_type::StreamposProxy	_posProxy;
	public:
		// conversion
		operator Item();
		// assignment
		void operator=(const Item& item);
	private:
		ItemProxy(IndexedFile* ifile, typename index_type::StreamposProxy posProxy);
	};

	class Iterator : public std::iterator<std::bidirectional_iterator_tag,
										  std::pair<const Key, ItemProxy> > {
		friend class IndexedFile;

		// data
		IndexedFile*					_ifile;
		typename index_type::iterator	_base;

	public:
		Iterator()			{}

		std::pair<const Key, ItemProxy>	operator*();
		Iterator&						operator++();
		Iterator						operator++(int);
		Iterator&						operator--();
		Iterator						operator--(int);
		typename index_type::iterator	base();

		bool operator==(const Iterator& x) const;
		bool operator!=(const Iterator& x) const;
	private:
		Iterator(IndexedFile* ifile, typename index_type::iterator base);
	};

	// construct / destroy
	explicit IndexedFile(Compare comp = Compare());
	explicit IndexedFile(const std::string& name, 
							std::ios_base::openmode mode = std::ios_base::in|std::ios_base::out, 
							Compare comp = Compare());
	~IndexedFile();

	// open/close
	bool			is_open();
	void			open(const std::string& name, std::ios_base::openmode mode = std::ios_base::in|std::ios_base::out);
	void			close();
	void			flush();

	// status check
	operator void*() const;
	bool operator!() const;

	// iterators
	Iterator		begin();
	Iterator		end();

	// capacity
	bool			empty();
	size_type		size();

	// modifiers
	Iterator		insert(const value_type& x);
	template<typename InputIterator>
	void			insert(InputIterator first, InputIterator last);

	void			erase(Iterator pos);
	size_type		erase(const Key& x);
	void			erase(Iterator first, Iterator last);
	void			clear();

	// observers
	Compare			key_comp() const;
	size_t			page_size()	const;
	index_type&		index();
	XDR_Stream&		xdrstream();

	// operations
	Iterator		find(const Key& x);
	size_type		count(const Key& x);

	Iterator		lower_bound(const Key& x);
	Iterator		upper_bound(const Key& x);
	std::pair<Iterator, Iterator>
					equal_range(const Key& x);

	// support for reference semantics and polymorphism
	Util::oXDR_Stream::ObjToIdMap&	obj2id();
	Util::oXDR_Stream::TypeToIdMap&	type2id();
	Util::iXDR_Stream::IdToObjMap&	id2obj();
	Util::iXDR_Stream::IdToTypeMap&	id2type();

	// utilities
	void			compact();
	Item			read(size_t pos);  // decodes and returns the Item at pos

private:
	// data
	index_type						_index;
	typename ItemTraits::xio_type	_xstream;

	// friendly utilities
	size_t			write(const Item& item); // returns the streampos of the written item
};


//
// IndexedFile::ItemProxy inline function definitions
//
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline 
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::ItemProxy::operator Item()
{ return _ifile->read(_posProxy); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::ItemProxy::operator=(const Item& item)
{ _posProxy = _ifile->write(item); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::ItemProxy::ItemProxy(
			IndexedFile* ifile,
			typename index_type::StreamposProxy posProxy)
  : _ifile(ifile), _posProxy(posProxy) {}


//
// IndexedFile::Iterator inline function definitions
//
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline std::pair<const Key, typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::ItemProxy>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::operator*()
{ 
	std::pair<const Key, typename index_type::StreamposProxy> rtn = *_base;
  	return std::make_pair( rtn.first, ItemProxy(_ifile, rtn.second) );
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::operator++()
{ ++_base; return *this; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::operator++(int)
{ Iterator t = *this; ++_base; return t; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::operator--()
{ --_base; return *this; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::operator--(int)
{ Iterator t = *this; --_base; return t; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::index_type::iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::base()
{ return _base; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::operator==(const Iterator& x) const
{ return _ifile == x._ifile and _base == x._base; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::operator!=(const Iterator& x) const
{ return !(*this == x); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline 
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator::Iterator(
			IndexedFile* ifile,
			typename index_type::iterator base)
  : _ifile(ifile), _base(base) {}


//
// IndexedFile inline function definitions
//
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::is_open()
{ return _index.is_open() and _xstream.is_open(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline 
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::operator void*() const
{
	if (!_index or !_xstream) return null;
	return const_cast<IndexedFile*>(this);
}

template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::operator!() const
{ return (!_index or !_xstream); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::begin()
{ return Iterator(this, _index.begin()); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::end()
{ return Iterator(this, _index.end()); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::empty()
{ return _index.empty(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::size_type
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::size()
{ return _index.size(); }


// Erase functions just erase the keys, the data becomes dead
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::erase(Iterator pos)
{ _index.erase(pos.base()); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::size_type
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::erase(const Key& k)
{ return _index.erase(k); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::erase(Iterator first, Iterator last)
{ _index.erase(first.base(), last.base()); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::clear()
{ _index.clear(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline Compare
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::key_comp() const
{ return _index.key_comp(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline size_t
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::page_size() const
{ return PageSize; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline BtreeIndex<Key,Compare,KeyTraits,PageSize>&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::index()
{ return _index; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline XDR_Stream&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::xdrstream()
{ return _xstream; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::find(const Key& k)
{ return Iterator(this, _index.find(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::size_type
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::count(const Key& k)
{ return _index.count(k); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::lower_bound(const Key& k)
{ return Iterator(this, _index.lower_bound(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::upper_bound(const Key& k)
{ return Iterator(this, _index.upper_bound(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline std::pair<typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator,
                 typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::equal_range(const Key& k)
{ return std::make_pair(lower_bound(k), upper_bound(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline Util::oXDR_Stream::ObjToIdMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::obj2id()
{ return xdrstream().obj2id(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline Util::oXDR_Stream::TypeToIdMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::type2id()
{ return xdrstream().type2id(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline Util::iXDR_Stream::IdToObjMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::id2obj()
{ return xdrstream().id2obj(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
inline Util::iXDR_Stream::IdToTypeMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::id2type()
{ return xdrstream().id2type(); }


} //> namespace Util


#ifdef EXPORT_NOT_SUPPORTED
#include "IndexedFile.hxx"
#endif


#endif

