/*******************************************************************************
  File: IndexedFile.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
	The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "IndexedFile.h"

namespace Util {


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::IndexedFile(Compare comp)
  : _index(comp),
	_xstream()
{
	// empty
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::IndexedFile(
			const std::string& name, 
			std::ios_base::openmode mode,
			Compare comp)
  : _index(comp),
    _xstream()
{
	open(name, mode);
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::~IndexedFile()
{
	if (is_open()) close();
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::open(
			const std::string& name,
			std::ios_base::openmode mode)
{
	const char* const func = "IndexedFile::open";

	// Add '.xd' to the name for the data file
	std::string dataName = name + ".xd";
	_xstream.open(dataName, mode);
	if (not _xstream.good()) 
		raise(std::ios_base::failure(
			make_what_str("ios_base::failure", "Could not open data file", func)));

	// Add '.xi' to the name for the index file
	std::string indexName = name + ".xi";
	_index.open(indexName, mode);
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::close()
{
	_index.close();
	_xstream.close();
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::flush()
{
	_index.flush();
	_xstream.flush();
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::insert(const value_type& x)
{
	// Write the data to the data file
	size_t link = write(x.second);

	// Add the key to the index
	typename index_type::iterator it = _index.insert( index_type::value_type(x.first, link) );

	return Iterator(this, it);
}



#ifndef NESTED_TEMPLATE_DEFINITION_NOT_SUPPORTED
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
template<typename InputIterator>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::insert(
			InputIterator first,
			InputIterator last)
{
	while (first != last) {
		insert(*first);
		++first;
	}
}
#endif


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::compact()
{
	// TBD
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
Item
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::read(size_t pos)
{
	Item item;
	_xstream.seekg(pos);
	ItemTraits::decode(_xstream, item);
	return item;
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits, size_t PageSize>
size_t
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits,PageSize>::write(const Item& item)
{
	_xstream.seekp(0, std::ios_base::end);
	size_t link = _xstream.tellp();
	ItemTraits::encode(_xstream, item);
	return link;
}


} //> namespace


